# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
This module defines an abstract base class, ArrayLike, which serves as an interface for objects that are
"array-like" in nature. The goal of this interface is to provide a set of methods that any class whether
it's a numpy.ndarray or any other specialized one (scipy's coo_array), can implement to ensure the
compatibility with certain behaviours expected for the SDK.
"""

from abc import ABC, abstractmethod
from typing import Any, Tuple, Union

import numpy as np
from scipy.sparse import coo_array


# Alias for the supported array types in the TitanQ SDK
Array = Union[np.ndarray, coo_array]


class ArrayLike(ABC):
    """
    This class represents an array like passed in by a user. We support different formats (e.g. sparse, dense),
    so this class will hide the specific format and allow different array formats to behave the same.
    """

    @abstractmethod
    def array(self) -> Any:
        """Returns the instance itself of the array like."""
        pass

    @abstractmethod
    def shape(self) -> Tuple:
        """Returns the shape of the array like."""
        pass


    @abstractmethod
    def np_type(self) -> np.dtype:
        """Returns the type of the array like in a np.dtype format."""
        pass


    @abstractmethod
    def to_bytes(self) -> bytes:
        """Returns the np type array as bytes."""
        pass



# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
The sparse threshold strategy allows to convert the arrays into others depending on a threshold.

This can help optimizing the chosen array, by helping the end-user to obtain much faster results by
opting for sparse arrays, or even by reducing his total RAM Usage.
"""

import numpy as np
from scipy.sparse import coo_array

from titanq._model.coord_store import CoordStore
from titanq._model.transform import (
    ArrayTransformStrategy,
    StrategyInputArrays,
    StrategyOutputArrays
)


_SPARSE_THRESHOLD = 0.2 # 20%


class SparseThresholdStrategy(ArrayTransformStrategy):
    """
    This strategy bases on the sparse threshold, meaning if the calculated density
    is abover certain threshold, the array would transform to another one.
    """

    def check_and_transform(self, array: StrategyInputArrays) -> StrategyOutputArrays:
        array_type_dispatcher = {
            CoordStore: self._handle_coordstore,
            coo_array: self._handle_coo_array,
            np.ndarray: self._handle_ndarray
        }

        handler = array_type_dispatcher.get(type(array))
        if handler:
            return handler(array)
        else:
            raise ValueError(f"Unsupported array type for the sparse threshold check: {type(array)}")


    def _handle_coordstore(self, array: CoordStore) -> StrategyOutputArrays:
        density = (len(array) + 1) / (array.problem_size() * array.problem_size())

        if density > _SPARSE_THRESHOLD:
            return array.to_numpy_array()
        else:
            return array.to_coo_array()


    def _handle_coo_array(self, array: coo_array) -> StrategyOutputArrays:
        density = array.nnz / array.shape[0] * array.shape[1]

        if density > _SPARSE_THRESHOLD:
            raise NotImplementedError("Sparse threshold strategy does not support yet conversion from coo_array to numpy arrays.")
        else:
            return array # do nothing


    def _handle_ndarray(self, array: np.ndarray) -> StrategyOutputArrays:
        density = np.count_nonzero(array) / array.size
        if density > _SPARSE_THRESHOLD:
            return array # do nothing
        else:
            return NotImplementedError("Sparse threshold strategy does not support yet conversion from numpy arrays to coo_array.")


# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
Array Transform module

This module provides functionality for checking various types of arrays and transform them in need.
It transforms them into different formats based on the chosen strategy.

It utilizes the strategy pattern, to provide a flexible approach for handling different array formats
and different check logic.

Example Usage
-------------

    # choose the strategy
    a_strategy = AStrategy()

    # create the transform checker with the strategy
    array_checker = ArrayTransformChecker(AStrategy)

    # create the array
    array = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 0]])

    # use the transform checker
    result_array = array_checker.check_and_transform(array)
"""

from abc import ABC, abstractmethod
from typing import Union

from titanq._model.array import Array
from titanq._model.coord_store import CoordStore


# accepted arrays and what the the module can output
StrategyInputArrays = Union[Array, CoordStore]
StrategyOutputArrays = Array


class ArrayTransformStrategy(ABC):
    """
    Abstract class that each strategy would implement.
    """

    @abstractmethod
    def check_and_transform(self, array: StrategyInputArrays) -> StrategyOutputArrays:
        """
        Checks the array and transform if needed. It can output the same array as the input.

        :param array: The array to check and transform if needed

        :returns: The transformed array depending on the chosen strategy
        """
        pass


class ArrayTransformChecker:
    """
    Array transform checker class that would apply a given strategy.

    You can apply the strategy at the creation time or by calling `set_strategy()`.
    """

    def __init__(self, strategy: ArrayTransformStrategy):
        self._strategy = strategy

    def check_and_transform(self, array: StrategyInputArrays) -> StrategyOutputArrays:
        return self._strategy.check_and_transform(array)


# Copyright (c) 2024, InfinityQ Technology, Inc.

from typing import Iterable, Tuple, Union

import numpy as np
from scipy.sparse import coo_array

CoordStoreType = Tuple[Tuple[int, int], Union[int, float]]


class CoordStore:
    """
    A simple and efficient coordinate store that allows adding or updating values
    at specific (x, y) coordinates.

    It uses a dictionary to store the coordinates, enabling O(1) lookup time for
    both updates and retrievals
    """

    def __init__(self, problem_size: int):
        """
        Initialize a CoordStore.

        :param size: A saved sized used to determine the total length of the storage, this
        is used for metadata only

        NOTE: The problem size does not have any correlation with the coordinates, append()
        method does not check if the coordinate is inside the size. It is simply used as a
        metadata for the class user.
        """
        self._problem_size = problem_size
        self._coords = {}


    def append(self, x: int, y: int, value: Union[int, float]) -> None:
        """Adds or update a value at the given (x, y) coordinate."""
        self._coords[(x, y)] = self._coords.get((x, y), 0) + value


    def problem_size(self) -> int:
        """Returns the problem size."""
        return self._problem_size


    def is_empty(self) -> bool:
        """Returns True if the the coordinate store is empty, otherwise False."""
        return not self._coords


    def __len__(self) -> int:
        """Returns the number of element in the store."""
        return len(self._coords)


    def __iter__(self) -> Iterable[CoordStoreType]:
        for key, value in self._coords.items():
            yield key, value


    def items(self) -> Iterable[CoordStoreType]:
        """Iterates over the coordinates with existing values."""
        return iter(self)


    def to_numpy_array(self) -> np.ndarray:
        """Converts into a dense numpy array."""
        array = np.zeros((self.problem_size(), self.problem_size()), dtype=np.float32)

        for (x, y), value in self.items():
            array[x, y] = value
        return array


    def to_coo_array(self) -> coo_array:
        """Converts into a sparse scipy COOrdinate matrix"""
        rows, cols = zip(*self._coords.keys())
        values = list(self._coords.values())

        return coo_array((values, (rows, cols)), shape=(self.problem_size(), self.problem_size()), dtype=np.float32)



# Copyright (c) 2025, InfinityQ Technology, Inc.

import io
from typing import Tuple

import numpy as np

from titanq._model.array import ArrayLike


class NumpyArray(ArrayLike):

    def __init__(self, array: np.ndarray):
        """
        Creates a NumpyArray implementing ArrayLike

        :param array: The numpy ndarray itself
        """
        self._array = array


    def array(self) -> np.ndarray:
        return self._array


    def shape(self) -> Tuple:
        return self.array().shape


    def np_type(self) -> np.dtype:
        return self.array().dtype


    def to_bytes(self) -> bytes:
        buffer = io.BytesIO()
        np.save(buffer, self.array())
        return buffer.getvalue()


# Copyright (c) 2025, InfinityQ Technology, Inc.

from typing import Optional

import numpy as np
from scipy.sparse import coo_array

from titanq._model.array import Array, ArrayLike
from titanq._model.array.numpy_array import NumpyArray
from titanq._model.array.scipy_coo_array import ScipyCooArray


class ArrayLikeFactory:
    """
    Factory class for the array like type.

    `create()` simply creates an ArrayLike based on the instance type
    """

    def create(self, array: Optional[Array]) -> Optional[ArrayLike]:
        """Returns the corresponding ArrayLike based on the instance type."""
        if array is None:
            return None

        if isinstance(array, np.ndarray):
            return NumpyArray(array)
        elif isinstance(array, coo_array):
            return ScipyCooArray(array)
        else:
            raise ValueError(f"Unsupported array type: {type(array)}")
